package java.nio.file.attribute;

import java.time.LocalDateTime;
import java.time.ZoneOffset;
import java.util.Objects;
import java.time.Instant;
import java.util.concurrent.TimeUnit;

public final class FileTime implements Comparable<FileTime>
{
    private final TimeUnit unit;
    private final long value;
    private Instant instant;
    private String valueAsString;
    private static final long HOURS_PER_DAY = 24L;
    private static final long MINUTES_PER_HOUR = 60L;
    private static final long SECONDS_PER_MINUTE = 60L;
    private static final long SECONDS_PER_HOUR = 3600L;
    private static final long SECONDS_PER_DAY = 86400L;
    private static final long MILLIS_PER_SECOND = 1000L;
    private static final long MICROS_PER_SECOND = 1000000L;
    private static final long NANOS_PER_SECOND = 1000000000L;
    private static final int NANOS_PER_MILLI = 1000000;
    private static final int NANOS_PER_MICRO = 1000;
    private static final long MIN_SECOND = -31557014167219200L;
    private static final long MAX_SECOND = 31556889864403199L;
    private static final long DAYS_PER_10000_YEARS = 3652425L;
    private static final long SECONDS_PER_10000_YEARS = 315569520000L;
    private static final long SECONDS_0000_TO_1970 = 62167219200L;
    
    private FileTime(long value, TimeUnit unit, Instant instant) {
        this.value = value;
        this.unit = unit;
        this.instant = instant;
    }
    
    public static FileTime from(long n, TimeUnit obj) {
        Objects.requireNonNull(obj, "unit");
        return new FileTime(n, obj, null);
    }
    
    public static FileTime fromMillis(long n) {
        return new FileTime(n, TimeUnit.MILLISECONDS, null);
    }
    
    public static FileTime from(Instant obj) {
        Objects.requireNonNull(obj, "instant");
        return new FileTime(0L, null, obj);
    }
    
    public long to(TimeUnit obj) {
        Objects.requireNonNull(obj, "unit");
        if (this.unit != null) {
            return obj.convert(this.value, this.unit);
        }
        long convert = obj.convert(this.instant.getEpochSecond(), TimeUnit.SECONDS);
        if (convert == Long.MIN_VALUE || convert == Long.MAX_VALUE) {
            return convert;
        }
        long convert2 = obj.convert((long)this.instant.getNano(), TimeUnit.NANOSECONDS);
        long n = convert + convert2;
        if (((convert ^ n) & (convert2 ^ n)) < 0L) {
            return (convert < 0L) ? Long.MIN_VALUE : Long.MAX_VALUE;
        }
        return n;
    }
    
    public long toMillis() {
        if (this.unit != null) {
            return this.unit.toMillis(this.value);
        }
        long epochSecond = this.instant.getEpochSecond();
        int nano = this.instant.getNano();
        long n = epochSecond * 1000L;
        if ((Math.abs(epochSecond) | 0x3E8L) >>> 31 != 0L && n / 1000L != epochSecond) {
            return (epochSecond < 0L) ? Long.MIN_VALUE : Long.MAX_VALUE;
        }
        return n + (long)(nano / 1000000);
    }
    
    private static long scale(long n, long n2, long n3) {
        if (n > n3) {
            return Long.MAX_VALUE;
        }
        if (n < -n3) {
            return Long.MIN_VALUE;
        }
        return n * n2;
    }
    
    public Instant toInstant() {
        if (this.instant == null) {
            int n = 0;
            long epochSecond;
            switch (this.unit) {
                case DAYS: {
                    epochSecond = scale(this.value, 86400L, 106751991167300L);
                    break;
                }
                case HOURS: {
                    epochSecond = scale(this.value, 3600L, 2562047788015215L);
                    break;
                }
                case MINUTES: {
                    epochSecond = scale(this.value, 60L, 153722867280912930L);
                    break;
                }
                case SECONDS: {
                    epochSecond = this.value;
                    break;
                }
                case MILLISECONDS: {
                    epochSecond = Math.floorDiv(this.value, 1000L);
                    n = (int)Math.floorMod(this.value, 1000L) * 1000000;
                    break;
                }
                case MICROSECONDS: {
                    epochSecond = Math.floorDiv(this.value, 1000000L);
                    n = (int)Math.floorMod(this.value, 1000000L) * 1000;
                    break;
                }
                case NANOSECONDS: {
                    epochSecond = Math.floorDiv(this.value, 1000000000L);
                    n = (int)Math.floorMod(this.value, 1000000000L);
                    break;
                }
                default: {
                    throw new AssertionError("Unit not handled");
                }
            }
            if (epochSecond <= -31557014167219200L) {
                this.instant = Instant.MIN;
            }
            else if (epochSecond >= 31556889864403199L) {
                this.instant = Instant.MAX;
            }
            else {
                this.instant = Instant.ofEpochSecond(epochSecond, (long)n);
            }
        }
        return this.instant;
    }
    
    @Override
    public boolean equals(Object o) {
        return o instanceof FileTime && this.compareTo((FileTime)o) == 0;
    }
    
    @Override
    public int hashCode() {
        return this.toInstant().hashCode();
    }
    
    private long toDays() {
        if (this.unit != null) {
            return this.unit.toDays(this.value);
        }
        return TimeUnit.SECONDS.toDays(this.toInstant().getEpochSecond());
    }
    
    private long toExcessNanos(long n) {
        if (this.unit != null) {
            return this.unit.toNanos(this.value - this.unit.convert(n, TimeUnit.DAYS));
        }
        return TimeUnit.SECONDS.toNanos(this.toInstant().getEpochSecond() - TimeUnit.DAYS.toSeconds(n));
    }
    
    @Override
    public int compareTo(FileTime fileTime) {
        if (this.unit != null && this.unit == fileTime.unit) {
            return Long.compare(this.value, fileTime.value);
        }
        long epochSecond = this.toInstant().getEpochSecond();
        int compare = Long.compare(epochSecond, fileTime.toInstant().getEpochSecond());
        if (compare != 0) {
            return compare;
        }
        int compare2 = Long.compare((long)this.toInstant().getNano(), (long)fileTime.toInstant().getNano());
        if (compare2 != 0) {
            return compare2;
        }
        if (epochSecond != 31556889864403199L && epochSecond != -31557014167219200L) {
            return 0;
        }
        long days = this.toDays();
        long days2 = fileTime.toDays();
        if (days == days2) {
            return Long.compare(this.toExcessNanos(days), fileTime.toExcessNanos(days2));
        }
        return Long.compare(days, days2);
    }
    
    private StringBuilder append(StringBuilder sb, int i, int n) {
        while (i > 0) {
            sb.append((char)(n / i + 48));
            n %= i;
            i /= 10;
        }
        return sb;
    }
    
    @Override
    public String toString() {
        if (this.valueAsString == null) {
            int nano = 0;
            long n;
            if (this.instant == null && this.unit.compareTo(TimeUnit.SECONDS) >= 0) {
                n = this.unit.toSeconds(this.value);
            }
            else {
                n = this.toInstant().getEpochSecond();
                nano = this.toInstant().getNano();
            }
            LocalDateTime localDateTime;
            int a;
            if (n >= -62167219200L) {
                long n4 = n - 315569520000L + 62167219200L;
                long n5 = Math.floorDiv(n4, 315569520000L) + 1L;
                localDateTime = LocalDateTime.ofEpochSecond(Math.floorMod(n4, 315569520000L) - 62167219200L, nano, ZoneOffset.UTC);
                a = localDateTime.getYear() + (int)n5 * 10000;
            }
            else {
                long n2 = n + 62167219200L;
                long n3 = n2 / 315569520000L;
                localDateTime = LocalDateTime.ofEpochSecond(n2 % 315569520000L - 62167219200L, nano, ZoneOffset.UTC);
                a = localDateTime.getYear() + (int)n3 * 10000;
            }
            if (a <= 0) {
                --a;
            }
            int nano2 = localDateTime.getNano();
            StringBuilder o = new StringBuilder(64);
            o.append((a < 0) ? "-" : "");
            int abs = Math.abs(a);
            if (abs < 10000) {
                this.append(o, 1000, Math.abs(abs));
            }
            else {
                o.append(String.valueOf(abs));
            }
            o.append('-');
            this.append(o, 10, localDateTime.getMonthValue());
            o.append('-');
            this.append(o, 10, localDateTime.getDayOfMonth());
            o.append('T');
            this.append(o, 10, localDateTime.getHour());
            o.append(':');
            this.append(o, 10, localDateTime.getMinute());
            o.append(':');
            this.append(o, 10, localDateTime.getSecond());
            if (nano2 != 0) {
                o.append('.');
                int n6;
                for (n6 = 100000000; nano2 % 10 == 0; nano2 /= 10, n6 /= 10) {}
                this.append(o, n6, nano2);
            }
            o.append('Z');
            this.valueAsString = o.toString();
        }
        return this.valueAsString;
    }
}
